c     -*- mode: FORTRAN -*-
c
c     This file is part of krot,
c     a program for the simulation, assignment and fit of HRLIF spectra.
c
c     Copyright (C) 1994-1998 Arnim Westphal
c     Copyright (C) 1997,1998 Jochen Kpper
c     
c     $Id: derivatives.F,v 1.5 1998/10/09 12:33:39 arnim Exp $


#include <arni.h>


      subroutine rrde( J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci, fcompl )
c     evaluate the rigid rotor derivatives for complex hamiltonia

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate
      integer        g, J
c     flag for complex eigenvectors
      integer        fcompl
      real*8         deriv( dmeval, npar )
      real*8         ham( dmham, dmham ), vecr( dmham, dmham ), veci( dmham, dmham )

      ARNIROT_LAUNCH ( "Launching rrde." )

c     no derivatives for J=0 block
      if (J.eq.0) return
      
c     calculate offset to current J block for correct derivative loading
      iept = J*J
c     loop to count the three derivatives
      do g = 1, 3
c        zero the work space
         call matzer( ham, dmham, iact )
c        for A constant deriv <i|Jz**2|i> matrix
         if( g .eq. 1 )
     *        call jz2( ham, dmham, J )
c        for B constant deriv <i|Jx**2|j> matrix
         if( g .eq. 2 )
     *        call jx2( ham, dmham, J )
c        for C constant deriv <i|Jy**2|j> matrix
         if( g .eq. 3 )
     *        call jy2( ham, dmham, J )

c        loop to count the states
         if ( fcompl .eq. 1 ) then
c           consider complex eigenvectors
            do istate = 1, iact
c              zero current derivative position
               deriv( istate + iept, g ) = 0.d0
c              diagonal contributions to <Jg**2> operator derivative; g = z, x, y
               call decalc( deriv( istate + iept, g ), istate, ham, vecr, veci, dmham, iact, 0, 0 )
c              two-off-diagonal contributions to <Jg**2> operator derivative; g = z, x, y
               call decalc( deriv( istate + iept, g ), istate, ham, vecr, veci, dmham, iact, 2, 0 )
            end do
         else
c           consider only real eigenvectors (standard rr, not axis switched)
            do istate = 1, iact
c              zero current derivative position
               deriv( istate + iept, g ) = 0.d0
c              diagonal contributions to <Jg**2> operator derivative; g = z, x, y
               call decalr( deriv( istate + iept, g ), istate, ham, vecr, dmham, iact, 0 )
c              two-off-diagonal contributions to <Jg**2> operator derivative; g = z, x, y
               call decalr( deriv( istate + iept, g ), istate, ham, vecr, dmham, iact, 2 )
            end do
         end if
      end do
      
      return
      end


c------------------------------------------------------------------------------
      subroutine jxde(J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci)
c     evaluate the Jx derivatives

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate, J
      real*8         deriv(dmeval,npar)
      real*8         ham(dmham,dmham), vecr(dmham,dmham), veci(dmham,dmham)

      ARNIROT_LAUNCH ( "Launching jxde." )

c     calculate offset for correct derivative loading
      iept = J*J
c     zero the work space
      call matzer( ham, dmham, iact )
c     fill <|Jx|> matrix Dx = 1
      call jx(ham, dmham, J, 1.d0)
      do istate = 1, iact, 1
c        initialize derivative at zero
         deriv(istate+iept,4) = 0.d0
c        eigenvector sum over Jx operator matrix in K basis
         call decalc( deriv(istate+iept,4), istate, ham, vecr, veci, dmham, iact, 1, 0 )
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine jyde(J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci)
c     evaluate the Jy derivatives

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate, J
      real*8         deriv(dmeval,npar)
      real*8         ham(dmham,dmham), vecr(dmham,dmham), veci(dmham,dmham)

      ARNIROT_LAUNCH ( "Launching jyde." )

c     calculate offset for correct derivative loading
      iept = J*J
c     zero the work space
      call matzer(ham, dmham, iact)
c     fill <|Jy|> matrix Dy = 1
      call jy(ham, dmham, J, 1.d0)
      do istate = 1, iact, 1
c        initialize derivative at zero
         deriv(istate+iept,5) = 0.d0
c        eigenvector sum over Jy operator matrix in K basis
         call decalc( deriv(istate+iept,5), istate, ham, vecr, veci, dmham, iact, 1, 1 )
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine jzde(J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci)
c     evaluate the Jz derivatives

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate, J
      real*8         deriv(dmeval,npar)
      real*8         ham(dmham,dmham), vecr(dmham,dmham), veci(dmham,dmham)

      ARNIROT_LAUNCH ( "Launching jzde." )

c     calculate offset for correct derivative loading
      iept = J*J
c     zero the work space
      call matzer(ham, dmham, iact)
c     fill <|Jz|> matrix Dz = 1
      call jz(ham, dmham, J, 1.d0)
      do istate = 1, iact, 1
c        initialize derivative at zero
         deriv(istate+iept,6) = 0.d0
c        eigenvector sum over Jz operator matrix in K basis
         call decalc( deriv(istate+iept,6), istate, ham, vecr, veci, dmham, iact, 0, 0 )
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine jzjxde(J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci)
c     evaluate the JzJx+JxJz derivatives

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate, J
      real*8         deriv(dmeval,npar)
      real*8         ham(dmham,dmham), vecr(dmham,dmham), veci(dmham,dmham)

      ARNIROT_LAUNCH ( "Launching jzjxde." )

c     calculate offset for correct derivative loading
      iept = J*J
c     zero the work space
      call matzer(ham, dmham, iact)
c     fill <|JzJx+JxJz|> matrix Dzx = 1
      call jzjx(ham, dmeval, J, 1.d0)
      do istate = 1, iact, 1
c        initialize derivative at zero
         deriv(istate+iept,5) = 0.d0
c        eigenvector sum over JzJx+JxJz operator matrix in K basis
         call decalc( deriv(istate+iept,5), istate, ham, vecr, veci, dmham, iact, 1, 0 )
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine jzjyde(J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci)
c     evaluate the JzJy+JyJz derivatives

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate, J
      real*8         deriv(dmeval,npar)
      real*8         ham(dmham,dmham), vecr(dmham,dmham), veci(dmham,dmham)

      ARNIROT_LAUNCH ( "Launching jzjyde." )

c     calculate offset for correct derivative loading
      iept = J*J
c     zero the work space
      call matzer(ham, dmham, iact)
c     fill <|JzJy+JyJz|> matrix Dzy = 1
      call jzjy(ham, dmeval, J, 1.d0)
      do istate = 1, iact, 1
c        initialize derivative at zero
         deriv(istate+iept,6) = 0.d0
c        eigenvector sum over JzJy+JyJz operator matrix in K basis
         call decalc( deriv(istate+iept,6), istate, ham, vecr, veci, dmham, iact, 1, 1 )
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine jyjxde(J, deriv, dmeval, npar, dmham, iact, ham, vecr, veci)
c     evaluate the JyJx+JxJy derivatives

      implicit none

      integer        dmham, dmeval, npar
      integer        iact, iept, istate, J
      real*8         deriv(dmeval,npar)
      real*8         ham(dmham,dmham), vecr(dmham,dmham), veci(dmham,dmham)

      ARNIROT_LAUNCH ( "Launching jyjxde." )

c     calculate offset for correct derivative loading
      iept = J*J
c     zero the work space
      call matzer(ham, dmham, iact)
c     fill <|JyJx+JxJy|> matrix Dyx = 1
      call jyjx(ham, dmeval, J, 1.d0)
      do istate = 1, iact, 1
c        initialize derivative at zero
         deriv(istate+iept,4) = 0.d0
c        eigenvector sum over JyJx+JxJy operator matrix in K basis
         call decalc( deriv(istate+iept,4), istate, ham, vecr, veci, dmham, iact, 2, 1 )
      end do

      return
      end


c------------------------------------------------------------------------------
      subroutine decalc( deriv, state, ham, vecr, veci, dim, act, ord, re0im1 )
c     generic routine to calculate derivatives for real or imaginary operators

c     deriv        = derivative being worked on
c     state        = number of state the derivative corresponds to
c     ham          = work space of operator matrix elements in K basis
c     vecr, veci   = real and imaginary eigenvectors for state
c     dim          = maximum (physical) dimension of matrices and vectors
c     act          = actual dimension of matrices and vectors
c     ord          = order of derivative summation off diagonal
c     re0im1       = flag for operator type, real(0) OR imaginary(1)

      implicit none

      integer        dim
      integer        act, state
      integer        ord, re0im1
      real*8         deriv
      real*8         ham( dim, dim ), vecr( dim, dim ), veci( dim, dim )

      integer        col, row

      ARNIROT_LAUNCH ( "Launching decalc." )

c     real operators
      if( re0im1 .eq. 0 ) then
c        off-diagonal by ord terms in summation row=col-ord
         if( ord .ne. 0 ) then
            do row = 1, act - ord
               col = row + ord
c              <i,j> real contribution ( <i,i> if ord == 0 )
               deriv = deriv + ham( row, col ) * vecr( row, state ) * vecr( col, state )
c              <i,j> imaginary contribution
               deriv = deriv + ham( row, col ) * veci( row, state ) * veci( col, state )
c              <j,i> contributions
c              real <j,i> contributions
               deriv = deriv + ham( col, row ) * vecr( col, state ) * vecr( row, state )
c              imaginary <j,i> contributions
               deriv = deriv + ham( col, row ) * veci( col, state ) * veci( row, state )
            end do
         else
c           diagonal elements only need to be counted once, only the first and second line (v.s.)
c           remain. Since ord == 0 => row == col, we can put them into one line:
            do row = 1, act - ord
               deriv = deriv + ham( row, row ) * ( vecr( row, state ) * vecr( row, state )
     *                                             + veci( row, state ) * veci( row, state ) )
            end do
         end if

c     imaginary operators
      else
         if( ord .ne. 0 ) then
            do row = 1, act-ord
               col = row + ord
c              <i,j> imaginary contributions
               deriv = deriv + ham( row, col ) * ( veci( row, state ) * vecr( col, state )
     *                                             - vecr( row, state ) * veci( col, state ) )
c              <j,i> imaginary contributions
               deriv = deriv + ham( col, row ) * ( veci( col, state ) * vecr( row, state )
     *                                             - vecr( col, state ) * veci( row, state ) )
            end do
c        else
c           prevent diagonal elements from being counted twice, need only <i,j> contributions
c           Here these cancel each other, we dont need them.
c           do row = 1, act - ord
c              col = row
c              deriv = deriv - ham( row, row ) * vecr( row, state ) * veci( row, state )
c              deriv = deriv + ham( row, row ) * veci( row, state ) * vecr( row, state )
c           end do
         end if
      end if
      
      return
      end


c------------------------------------------------------------------------------
      subroutine decalr( deriv, state, ham, vecr, dim, act, ord )
c     generic routine to calculate derivatives for real or imaginary operators

c     deriv        = derivative being worked on
c     state        = number of state the derivative corresponds to
c     ham          = work space of operator matrix elements in K basis
c     vecr         = real eigenvectors for state
c     dim          = maximum (physical) dimension of matrices and vectors
c     act          = actual dimension of matrices and vectors
c     ord          = order of derivative summation off diagonal

      implicit none

      integer        dim
      integer        act, ord, state
      real*8         deriv
      real*8         ham( dim, dim ), vecr( dim, dim )

      integer        col, row

      ARNIROT_LAUNCH ( "Launching decalr." )

c     reduced calculation expense for purely real operators and eigenvectors
c     off-diagonal by ord terms in summation row=col-ord
      if( ord .ne. 0 ) then
         do row = 1, act - ord
            col = row + ord
c           <i,j> and <j,i> real contributions
            deriv = deriv + ( ham( row, col ) + ham( col, row ) )
     *                                      * vecr( row, state ) * vecr( col, state )
         end do
      else
c        if ord==0 => row==col, only ham element <i,i> is considered:
         do row = 1, act
            deriv = deriv + ham( row, row ) * vecr( row, state ) * vecr( row, state )
         end do
      end if
      
      return
      end
